"""
The Grid module contains several utilities for grid operations
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import re

from DIRAC.Core.Utilities.Os import sourceEnv
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ConfigurationSystem.Client.Helpers import Local
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities.Subprocess import systemCall, shellCall

__RCSID__ = "$Id$"


def executeGridCommand(proxy, cmd, gridEnvScript=None):
    """
    Execute cmd tuple after sourcing GridEnv
    """
    currentEnv = dict(os.environ)

    if not gridEnvScript:
        # if not passed as argument, use default from CS Helpers
        gridEnvScript = Local.gridEnv()

    if gridEnvScript:
        command = gridEnvScript.split()
        ret = sourceEnv(10, command)
        if not ret["OK"]:
            return S_ERROR("Failed sourcing GridEnv: %s" % ret["Message"])
        gridEnv = ret["outputEnv"]
        #
        # Preserve some current settings if they are there
        #
        if "X509_VOMS_DIR" in currentEnv:
            gridEnv["X509_VOMS_DIR"] = currentEnv["X509_VOMS_DIR"]
        if "X509_CERT_DIR" in currentEnv:
            gridEnv["X509_CERT_DIR"] = currentEnv["X509_CERT_DIR"]
    else:
        gridEnv = currentEnv

    if not proxy:
        res = getProxyInfo()
        if not res["OK"]:
            return res
        gridEnv["X509_USER_PROXY"] = res["Value"]["path"]
    elif isinstance(proxy, six.string_types):
        if os.path.exists(proxy):
            gridEnv["X509_USER_PROXY"] = proxy
        else:
            return S_ERROR("Can not treat proxy passed as a string")
    else:
        ret = gProxyManager.dumpProxyToFile(proxy)
        if not ret["OK"]:
            return ret
        gridEnv["X509_USER_PROXY"] = ret["Value"]

    result = systemCall(120, cmd, env=gridEnv)
    return result


def ldapsearchBDII(filt=None, attr=None, host=None, base=None, selectionString="Glue"):
    """Python wrapper for ldapserch at bdii.

    :param  filt: Filter used to search ldap, default = '', means select all
    :param  attr: Attributes returned by ldapsearch, default = '*', means return all
    :param  host: Host used for ldapsearch, default = 'cclcgtopbdii01.in2p3.fr:2170', can be changed by $LCG_GFAL_INFOSYS

    :return: standard DIRAC answer with Value equals to list of ldapsearch responses

    Each element of list is dictionary with keys:

      'dn':                 Distinguished name of ldapsearch response
      'objectClass':        List of classes in response
      'attr':               Dictionary of attributes
    """

    if filt is None:
        filt = ""
    if attr is None:
        attr = ""
    if host is None:
        host = "cclcgtopbdii01.in2p3.fr:2170"
    if base is None:
        base = "Mds-Vo-name=local,o=grid"

    if isinstance(attr, list):
        attr = " ".join(attr)

    cmd = 'ldapsearch -x -LLL -o ldif-wrap=no -h %s -b %s "%s" %s' % (host, base, filt, attr)
    result = shellCall(0, cmd)

    response = []

    if not result["OK"]:
        return result

    status = result["Value"][0]
    stdout = result["Value"][1]
    stderr = result["Value"][2]

    if status != 0:
        return S_ERROR(stderr)

    lines = []
    for line in stdout.split("\n"):
        if line.find(" ") == 0:
            lines[-1] += line.strip()
        else:
            lines.append(line.strip())

    record = None
    for line in lines:
        if line.find("dn:") == 0:
            record = {
                "dn": line.replace("dn:", "").strip(),
                "objectClass": [],
                "attr": {"dn": line.replace("dn:", "").strip()},
            }
            response.append(record)
            continue
        if record:
            if line.find("objectClass:") == 0:
                record["objectClass"].append(line.replace("objectClass:", "").strip())
                continue
            if line.find(selectionString) == 0:
                index = line.find(":")
                if index > 0:
                    attr = line[:index]
                    value = line[index + 1 :].strip()
                    if attr in record["attr"]:
                        if isinstance(record["attr"][attr], list):
                            record["attr"][attr].append(value)
                        else:
                            record["attr"][attr] = [record["attr"][attr], value]
                    else:
                        record["attr"][attr] = value

    return S_OK(response)



########################################################################
# File: File.py
# Date: 2012/08/03 15:02:53
########################################################################
"""
:mod: File

.. module: File
  :synopsis: RMS operation file

operation file
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Disable invalid names warning
# pylint: disable=invalid-name

__RCSID__ = "$Id$"

# # imports
import datetime
import os
import json
import six

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.File import checkGuid
from DIRAC.RequestManagementSystem.private.JSONUtils import RMSEncoder


########################################################################
class File(object):

    """
     A bag object holding Operation file attributes.

    :param Operation.Operation _parent: reference to parent Operation
    :param dict __data__: attrs dict


    It is managed by SQLAlchemy, so the OperationID, FileID should never be set by hand
    (except when constructed from JSON of course...)
    In principle, the _parent attribute could be totally managed by SQLAlchemy. However, it is
    set only when inserted into the DB, this is why I manually set it in the Operation

    .. warning::

      You cannot add a File object to multiple Operations. They are different entry in the DB, so they must be different
      objects

    """

    _datetimeFormat = "%Y-%m-%d %H:%M:%S"

    def __init__(self, fromDict=None):
        """c'tor

        :param self: self reference
        :param dict fromDict: property dict
        """
        self._parent = None
        self._Status = "Waiting"
        self._LFN = None
        self.PFN = None
        self._ChecksumType = None
        self.Checksum = None
        self._GUID = None
        self.Attempt = 0
        self.Size = 0
        self.Error = None
        self._duration = 0

        # This variable is used in the setter to know whether they are called
        # because of the json initialization or not
        self.initialLoading = True

        fromDict = (
            fromDict
            if isinstance(fromDict, dict)
            else json.loads(fromDict)
            if isinstance(fromDict, six.string_types)
            else {}
        )

        for attrName, attrValue in fromDict.items():
            # The JSON module forces the use of UTF-8, which is not properly
            # taken into account in DIRAC.
            # One would need to replace all the '== str' with 'in six.string_types'
            # This is converting `unicode` to `str` and doesn't make sense in Python 3
            if six.PY2 and isinstance(attrValue, six.string_types):
                attrValue = attrValue.encode()
            if attrValue:
                setattr(self, attrName, attrValue)

        self.initialLoading = False

    @property
    def LFN(self):
        """LFN prop"""
        return self._LFN

    @LFN.setter
    def LFN(self, value):
        """lfn setter"""
        if not isinstance(value, six.string_types):
            raise TypeError("LFN has to be a string!")
        if not os.path.isabs(value):
            raise ValueError("LFN should be an absolute path!")
        self._LFN = value

    @property
    def GUID(self):
        """GUID prop"""
        return self._GUID

    @GUID.setter
    def GUID(self, value):
        """GUID setter"""
        if value:
            if not isinstance(value, six.string_types):
                raise TypeError("GUID should be a string!")
            if not checkGuid(value):
                raise ValueError("'%s' is not a valid GUID!" % str(value))
        self._GUID = value

    @property
    def ChecksumType(self):
        """checksum type prop"""
        return self._ChecksumType

    @ChecksumType.setter
    def ChecksumType(self, value):
        """checksum type setter"""
        if not value:
            self._ChecksumType = ""
        elif value and str(value).strip().upper() not in ("ADLER32", "MD5", "SHA1"):
            if str(value).strip().upper() == "AD":
                self._ChecksumType = "ADLER32"
            else:
                raise ValueError("unknown checksum type: %s" % value)
        else:
            self._ChecksumType = str(value).strip().upper()

    @property
    def Status(self):
        """status prop"""
        if not self._Status:
            self._Status = "Waiting"
        return self._Status

    @Status.setter
    def Status(self, value):
        """status setter"""
        if value not in ("Waiting", "Failed", "Done", "Scheduled"):
            raise ValueError("Unknown Status: %s!" % str(value))

        if value == "Done":
            self.Error = ""

        updateTime = self._Status != value
        if updateTime and self._parent:
            self._parent.LastUpdate = datetime.datetime.utcnow().replace(microsecond=0)

        self._Status = value

        if self._parent:
            self._parent._notify()

    def __str__(self):
        """str operator"""
        return self.toJSON()["Value"]

    def toJSON(self):
        """Returns the json formated string that describes the File"""
        try:
            jsonStr = json.dumps(self, cls=RMSEncoder)
            return S_OK(jsonStr)
        except Exception as e:
            return S_ERROR(str(e))

    def _getJSONData(self):
        """Returns the data that have to be serialized by JSON"""
        attrNames = [
            "FileID",
            "OperationID",
            "Status",
            "LFN",
            "PFN",
            "ChecksumType",
            "Checksum",
            "GUID",
            "Attempt",
            "Size",
            "Error",
        ]

        jsonData = {}

        for attrName in attrNames:

            # FileID and OperationID might not be set since they are managed by SQLAlchemy
            if not hasattr(self, attrName):
                continue

            jsonData[attrName] = getattr(self, attrName)
            value = getattr(self, attrName)

            if isinstance(value, datetime.datetime):
                # We convert date time to a string
                jsonData[attrName] = value.strftime(self._datetimeFormat)  # pylint: disable=no-member
            else:
                jsonData[attrName] = value

        return jsonData



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-pilot-logging-info.py
# Author :  Stuart Paterson
########################################################################
"""
Retrieve logging info of a Grid pilot

Example:
  $ dirac-admin-get-pilot-logging-info https://marlb.in2p3.fr:9000/26KCLKBFtxXKHF4_ZrQjkw
  Pilot Reference: dirac-admin-get-pilot-logging-info https://marlb.in2p3.fr:9000/26KCLKBFtxXKHF4_ZrQjkw
  ===================== glite-job-logging-info Success =====================
  LOGGING INFORMATION:
  Printing info for the Job : https://marlb.in2p3.fr:9000/26KCLKBFtxXKHF4_ZrQjkw
      ---
  Event: RegJob
  - Arrived   =  Mon Feb 21 13:27:50 2011 CET
  - Host      =  marwms.in2p3.fr
  - Jobtype   =  SIMPLE
  - Level     =  SYSTEM
  - Ns        =  https://marwms.in2p3.fr:7443/glite_wms_wmproxy_server
  - Nsubjobs  =  0
  - Parent    =  https://marlb.in2p3.fr:9000/WQHVOB1mI4oqrlYz2ZKtgA
  - Priority  =  asynchronous
  - Seqcode   =  UI=000000:NS=0000000001:WM=000000:BH=0000000000:JSS=000000:LM=000000:LRMS=000000:APP=000000:LBS=000000
  - Source    =  NetworkServer
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

# pylint: disable=wrong-import-position
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["PilotID:  Grid ID of the pilot"])
    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    for gridID in args:

        result = diracAdmin.getPilotLoggingInfo(gridID)
        if not result["OK"]:
            errorList.append((gridID, result["Message"]))
            exitCode = 2
        else:
            print("Pilot Reference: %s", gridID)
            print(result["Value"])
            print()

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



""" Test class for QueueUtilities
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest
import copy
from mock import MagicMock


from DIRAC import S_OK
from DIRAC.WorkloadManagementSystem.Utilities.QueueUtilities import *


siteDict1 = {
    "Site1": {
        "CE1": {"Queues": {"Queue1": {}, "Queue2": {}}, "CEType": "Type1"},
        "CE2": {"Queues": {"Queue1": {}}, "CEType": "Type2"},
    },
    "Site2": {"CE3": {"Queues": {"Queue1": {"NumberOfProcessors": 3}}, "CEType": "Type2"}},
}


expectedQueueDict1 = {
    "CE1_Queue1": {
        "CEName": "CE1",
        "CEType": "Type1",
        "QueryCEFlag": False,
        "QueueName": "Queue1",
        "Site": "Site1",
        "ParametersDict": {
            "GridCE": "CE1",
            "GridEnv": None,
            "Queue": "Queue1",
            "RequiredTag": [],
            "Setup": None,
            "Site": "Site1",
            "Tag": [],
            "WorkingDirectory": "Queue1",
        },
    },
    "CE1_Queue2": {
        "CEName": "CE1",
        "CEType": "Type1",
        "QueryCEFlag": False,
        "QueueName": "Queue2",
        "Site": "Site1",
        "ParametersDict": {
            "GridCE": "CE1",
            "GridEnv": None,
            "Queue": "Queue2",
            "RequiredTag": [],
            "Setup": None,
            "Site": "Site1",
            "Tag": [],
            "WorkingDirectory": "Queue2",
        },
    },
    "CE2_Queue1": {
        "CEName": "CE2",
        "CEType": "Type2",
        "QueryCEFlag": False,
        "QueueName": "Queue1",
        "Site": "Site1",
        "ParametersDict": {
            "GridCE": "CE2",
            "GridEnv": None,
            "Queue": "Queue1",
            "RequiredTag": [],
            "Setup": None,
            "Site": "Site1",
            "Tag": [],
            "WorkingDirectory": "Queue1",
        },
    },
    "CE3_Queue1": {
        "CEName": "CE3",
        "CEType": "Type2",
        "QueryCEFlag": False,
        "QueueName": "Queue1",
        "Site": "Site2",
        "ParametersDict": {
            "GridCE": "CE3",
            "GridEnv": None,
            "Queue": "Queue1",
            "RequiredTag": [],
            "Setup": None,
            "Site": "Site1",
            "Tag": ["MultiProcessor"],
            "NumberOfProcessors": 3,
            "WorkingDirectory": "Queue1",
        },
    },
}


@pytest.mark.parametrize(
    "queueDict, dictExpected",
    [
        ({}, {}),
        ({"notUsefulParam": ""}, {"notUsefulParam": ""}),
        ({"maxCPUTime": 45000}, {"maxCPUTime": 45000}),
        ({"SI00": 54}, {"SI00": 54}),
        ({"maxCPUTime": 45000, "SI00": 54}, {"maxCPUTime": 45000, "SI00": 54, "CPUTime": 583200}),
    ],
)
def test_computeQueueCPULimit(queueDict, dictExpected):
    """Test the computeCPULimit function"""
    computeQueueCPULimit(queueDict)
    assert queueDict == dictExpected


@pytest.mark.parametrize(
    "ceDict, queueDict, dictExpected",
    [
        ({}, {}, {"Tag": [], "RequiredTag": []}),
        ({"notUsefulParam": ""}, {"notUsefulParam": ""}, {"Tag": [], "RequiredTag": []}),
        ({"Tag": "Test"}, {}, {"Tag": ["Test"], "RequiredTag": []}),
        ({}, {"Tag": "Test"}, {"Tag": ["Test"], "RequiredTag": []}),
        ({"Tag": "Test"}, {"Tag": "Test"}, {"Tag": ["Test"], "RequiredTag": []}),
        ({"Tag": "Test1"}, {"Tag": "Test2"}, {"Tag": ["Test1", "Test2"], "RequiredTag": []}),
        ({"Tag": ["Test1", "Test2"]}, {"Tag": ["Test2", "Test2"]}, {"Tag": ["Test1", "Test2"], "RequiredTag": []}),
        ({"RequiredTag": "Test"}, {}, {"Tag": [], "RequiredTag": ["Test"]}),
        ({}, {"RequiredTag": "Test"}, {"Tag": [], "RequiredTag": ["Test"]}),
        ({"RequiredTag": "Test"}, {"RequiredTag": "Test"}, {"Tag": [], "RequiredTag": ["Test"]}),
        ({"RequiredTag": "Test1"}, {"RequiredTag": "Test2"}, {"Tag": [], "RequiredTag": ["Test1", "Test2"]}),
        (
            {"Tag": ["Test1", "Test2"], "RequiredTag": ["Test2"]},
            {"Tag": ["Test1", "Test3"], "RequiredTag": ["Test2", "Test3"]},
            {"Tag": ["Test1", "Test2", "Test3"], "RequiredTag": ["Test2", "Test3"]},
        ),
    ],
)
def test_resolveTags(ceDict, queueDict, dictExpected):
    """Test the resolveTags function"""
    resolveTags(ceDict, queueDict)
    assert queueDict["Tag"].sort() == dictExpected["Tag"].sort()
    assert queueDict["RequiredTag"].sort() == dictExpected["RequiredTag"].sort()


@pytest.mark.parametrize(
    "ceDict, queueDict, dictExpected",
    [
        ({}, {}, {}),
        ({"Platform": "x86_64"}, {}, {"Platform": "x86_64"}),
        ({"Platform": "notexist"}, {}, {"Platform": "notexist"}),
        ({}, {"Platform": "x86_64"}, {"Platform": "x86_64"}),
        ({}, {"Platform": "notexist"}, {"Platform": "notexist"}),
        ({"Platform": "x86_64", "OS": "centos"}, {}, {"Platform": "x86_64"}),
        ({"Platform": "notexist", "OS": "centos"}, {}, {"Platform": "notexist"}),
        ({}, {"Platform": "x86_64", "OS": "centos"}, {"Platform": "x86_64", "OS": "centos"}),
        ({}, {"Platform": "notexist", "OS": "centos"}, {"Platform": "notexist", "OS": "centos"}),
    ],
)
def test_setPlatform(ceDict, queueDict, dictExpected):
    """Test the setPlatform function"""
    setPlatform(ceDict, queueDict)
    assert queueDict == dictExpected


@pytest.mark.parametrize(
    "queueDict, queuesExpected",
    [
        (siteDict1, expectedQueueDict1),
    ],
)
def test_getQueuesResolved(mocker, queueDict, queuesExpected):
    """Test the getQueuesResolvedEnhanced function"""
    queueCECache = {}
    queueDictLocal = copy.deepcopy(queueDict)

    ce = MagicMock()
    ce.isValid = MagicMock(return_value=S_OK())
    ceFactoryModule = "DIRAC.Resources.Computing.ComputingElementFactory"
    ceFactoryMock = mocker.patch("%s.ComputingElementFactory.getCE" % ceFactoryModule, return_value=S_OK(ce))
    with ceFactoryMock:
        queueDictResolved = getQueuesResolved(queueDictLocal, queueCECache)

    assert queueDictResolved["OK"]
    for qName, qDictResolved in queueDictResolved["Value"].items():
        assert sorted(qDictResolved) == sorted(queuesExpected[qName])



""" DIRAC Multi VO FileCatalog plugin class to manage file metadata for multiple VO.
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.FileMetadata.FileMetadata import FileMetadata
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryMetadata.MultiVODirectoryMetadata import (
    _getMetaName,
    _getMetaNameDict,
    _stripSuffix,
)


class MultiVOFileMetadata(FileMetadata):
    """
    Multi-VO FileCatalog plugin implementation.
    """

    def __init__(self, database=None):
        super(MultiVOFileMetadata, self).__init__(database=database)

    def addMetadataField(self, pName, pType, credDict):
        """
        Add a new metadata parameter to the Metadata Database.
        Modified to use fully qualified metadata names.

        :param str pName: parameter name
        :param str pType: parameter type in the MySQL notation
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """

        fname = _getMetaName(pName, credDict)
        return super(MultiVOFileMetadata, self).addMetadataField(fname, pType, credDict)

    def deleteMetadataField(self, pName, credDict):
        """Remove metadata field.
        Table name is now fully qualified

        :param str pName: parameter name
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """
        fname = _getMetaName(pName, credDict)
        return super(MultiVOFileMetadata, self).deleteMetadataField(fname, credDict)

    def getFileMetadataFields(self, credDict):
        """Get all the defined metadata fields

        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """
        result = super(MultiVOFileMetadata, self)._getFileMetadataFields(credDict)
        if not result["OK"]:
            return result

        metaDict = _stripSuffix(result["Value"], credDict)
        return S_OK(metaDict)

    def setMetadata(self, dPath, metaDict, credDict):
        """Set the value of a given metadata field for the the given directory path

        :param str dPath: directory path
        :param dict metaDict: dictionary with the user metadata
        :param dict credDict: client credential dictionary

        :return: standard Dirac result object
        """
        fMetaDict = _getMetaNameDict(metaDict, credDict)
        return super(MultiVOFileMetadata, self).setMetadata(dPath, fMetaDict, credDict)

    def removeMetadata(self, dPath, metaList, credDict):
        """
        Remove the specified metadata for the given directory for users own VO.

        :param str dPath: directory path
        :param dict metaList: metadata names list
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """

        metaList = [_getMetaName(meta, credDict) for meta in metaList]
        result = super(MultiVOFileMetadata, self).removeMetadata(dPath, metaList, credDict)
        if not result["OK"]:
            if "FailedMetadata" in result:
                failedDict = _stripSuffix(result["FailedMetadata"], credDict)
                result["FailedMetadata"] = failedDict
            return result

        return S_OK()

    def setFileMetaParameter(self, path, metaName, metaValue, credDict):
        """
        Set an meta parameter - metadata which is not used in the the data
        search operations.

        :param str path: directory path
        :param str metaName: metadata name
        :param str metaValue: metadata value
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """
        fName = _getMetaName(metaName, credDict)
        return super(MultiVOFileMetadata, self).setFileMetaParameter(path, fName, metaValue, credDict)

    def getFileUserMetadata(self, path, credDict):
        """
        Get metadata for the given file.

        :param str path:  file path
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """

        result = super(MultiVOFileMetadata, self).getFileUserMetadata(path, credDict)
        if not result["OK"]:
            return result

        result["Value"] = _stripSuffix(result["Value"], credDict)
        return result

    def findFilesByMetadata(self, metaDict, path, credDict):
        """Find Files satisfying the given metadata

        :param dict metaDict: dictionary with the metaquery parameters
        :param str path: Path to search into
        :param dict credDict: Dictionary with the user credentials

        :return: S_OK/S_ERROR, Value ID:LFN dictionary of selected files
        """

        fMetaDict = _getMetaNameDict(metaDict, credDict)
        return super(MultiVOFileMetadata, self).findFilesByMetadata(fMetaDict, path, credDict)



"""Tests for the CheckMigration Operation"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

import pytest
from mock import MagicMock

from DIRAC import S_OK
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.Request import Request

from DIRAC.DataManagementSystem.Agent.RequestOperations import CheckMigration

MODULE = "DIRAC.DataManagementSystem.Agent.RequestOperations.CheckMigration"
FILE_NAME = "fileName"
N_FILES = 3


@pytest.fixture
def listOfLFNs():
    lfns = []
    for index, name in enumerate([FILE_NAME] * N_FILES):
        lfns.append("/vo/%s_%d" % (name, index))
    return lfns


@pytest.fixture
def seMock(mocker):
    """Mock call to StorageElement."""
    seModMock = mocker.MagicMock(name="StorageElementModule")
    seClassMock = mocker.MagicMock(name="StorageElementClass")
    seClassMock.getFileMetadata = mocker.MagicMock(return_value=S_OK({"Migrated": 0}))
    seModMock.return_value = seClassMock
    mocker.patch(MODULE + ".StorageElement", new=seModMock)
    return seModMock, seClassMock


@pytest.fixture
def checkRequestAndOp(listOfLFNs):
    req = Request()
    req.RequestName = "MyRequest"
    op = Operation()
    op.Type = "CheckMigration"
    for index, lfn in enumerate(listOfLFNs):
        oFile = File()
        oFile.LFN = lfn
        oFile.Size = index
        oFile.Checksum = "01130a%0d" % index
        oFile.ChecksumType = "adler32"
        op.addFile(oFile)
    req.addOperation(op)
    return req, op


@pytest.fixture
def multiRetVal(listOfLFNs):
    """Return a return structure for multiple values"""

    def retFunc(*args, **kwargs):
        retVal = {
            "OK": True,
            "Value": {
                "Failed": {},
                "Successful": {},
            },
        }
        for lfn in listOfLFNs:
            if kwargs.get("OK", not kwargs.get("Error", False)):
                retVal["Value"]["Successful"][lfn] = {"Migrated": kwargs.get("Migrated", 0)}
            else:
                retVal["Value"]["Failed"][lfn] = kwargs.get("Error", "Failed to do X")
        return retVal

    return retFunc


@pytest.fixture
def checkMigration(mocker, checkRequestAndOp):
    cm = CheckMigration.CheckMigration(checkRequestAndOp[1])
    return cm


def test_constructor(checkMigration):
    assert checkMigration.waitingFiles == []


def test_run_NotMigrated(checkMigration, seMock, multiRetVal):
    seModMock, seClassMock = seMock
    seClassMock.getFileMetadata = MagicMock(side_effect=functools.partial(multiRetVal, Migrated=0))
    checkMigration._run()
    assert len(checkMigration.waitingFiles) == N_FILES
    seModMock.assert_called_with("")
    for opFile in checkMigration.operation:
        assert opFile.Status == "Waiting"


def test_run_Migrated(checkMigration, seMock, multiRetVal):
    seModMock, seClassMock = seMock
    seClassMock.getFileMetadata = MagicMock(side_effect=functools.partial(multiRetVal, Migrated=1))
    checkMigration._run()
    assert len(checkMigration.waitingFiles) == N_FILES
    for opFile in checkMigration.operation:
        assert opFile.Status == "Done"


def test_run_Failed(checkMigration, seMock, multiRetVal):
    seModMock, seClassMock = seMock
    seClassMock.getFileMetadata = MagicMock(side_effect=functools.partial(multiRetVal, Error="Fail Fail Fail"))
    checkMigration._run()
    assert len(checkMigration.waitingFiles) == N_FILES
    for opFile in checkMigration.operation:
        assert opFile.Status == "Waiting"


def test_call_Migrated(checkMigration, seMock, multiRetVal):
    seModMock, seClassMock = seMock
    seClassMock.getFileMetadata = MagicMock(side_effect=functools.partial(multiRetVal, Migrated=1))
    assert checkMigration()["OK"]
    assert len(checkMigration.waitingFiles) == N_FILES
    for opFile in checkMigration.operation:
        assert opFile.Status == "Done"


def test_call_Exception(checkMigration, seMock, multiRetVal):
    seModMock, seClassMock = seMock
    seClassMock.getFileMetadata = MagicMock(side_effect=RuntimeError("Throw Down"))
    ret = checkMigration()
    assert not ret["OK"]
    assert ret["Message"] == "Throw Down"
    assert len(checkMigration.waitingFiles) == N_FILES
    for opFile in checkMigration.operation:
        assert opFile.Status == "Waiting"



""" Unit test for ConsistencyInspector
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import datetime
from mock import MagicMock

from DIRAC import gLogger
from DIRAC.Resources.Catalog.test.mock_FC import fc_mock

# sut
from DIRAC.DataManagementSystem.Client.ConsistencyInspector import ConsistencyInspector


class UtilitiesTestCase(unittest.TestCase):
    def setUp(self):

        gLogger.setLevel("DEBUG")

        self.lfnDict = {
            "aa.raw": {
                "aa.raw": {"FileType": "RAW", "RunNumber": 97019},
                "/lhcb/1_2_1.Semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
            "cc.raw": {
                "cc.raw": {"FileType": "RAW", "RunNumber": 97019},
                "/lhcb/1_1.semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
        }

        dmMock = MagicMock()
        dicMock = MagicMock()

        self.ci = ConsistencyInspector(transClient=MagicMock(), dm=dmMock, fc=fc_mock, dic=dicMock)
        self.ci.fileType = ["SEMILEPTONIC.DST", "LOG", "RAW"]
        self.ci.fileTypesExcluded = ["LOG"]
        self.ci.prod = 0
        self.maxDiff = None


class ConsistencyInspectorSuccess(UtilitiesTestCase):
    def test_getReplicasPresence(self):
        res = self.ci.getReplicasPresence(["/this/is/dir1/file1.txt", "/this/is/dir1/file2.foo.bar"])
        self.assertEqual(tuple(map(set, res)), ({"/this/is/dir1/file1.txt", "/this/is/dir1/file2.foo.bar"}, set()))

    def test__selectByFileType(self):
        lfnDict = {
            "aa.raw": {
                "bb.raw": {"FileType": "RAW", "RunNumber": 97019},
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "DST"},
                "/lhcb/1_2_1.Semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
            "cc.raw": {
                "dd.raw": {"FileType": "RAW", "RunNumber": 97019},
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "LOG"},
                "/lhcb/1_1.semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
        }

        res = self.ci._selectByFileType(lfnDict)

        lfnDictExpected = {
            "aa.raw": {
                "/lhcb/1_2_1.Semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
                "bb.raw": {"RunNumber": 97019, "FileType": "RAW"},
            },
            "cc.raw": {
                "dd.raw": {"RunNumber": 97019, "FileType": "RAW"},
                "/lhcb/1_1.semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
        }
        self.assertEqual(res, lfnDictExpected)

        lfnDict = {"aa.raw": {"/bb/pippo/aa.dst": {"FileType": "LOG"}, "bb.log": {"FileType": "LOG"}}}
        res = self.ci._selectByFileType(lfnDict)
        lfnDictExpected = {}
        self.assertEqual(res, lfnDictExpected)

    def test__getFileTypesCount(self):
        lfnDict = {"aa.raw": {"bb.log": {"FileType": "LOG"}, "/bb/pippo/aa.dst": {"FileType": "DST"}}}
        res = self.ci._getFileTypesCount(lfnDict)
        resExpected = {"aa.raw": {"DST": 1, "LOG": 1}}
        self.assertEqual(res, resExpected)

        lfnDict = {
            "aa.raw": {
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "DST"},
                "/bb/pippo/cc.dst": {"FileType": "DST"},
            }
        }
        res = self.ci._getFileTypesCount(lfnDict)
        resExpected = {"aa.raw": {"DST": 2, "LOG": 1}}
        self.assertEqual(res, resExpected)

    # def test__catalogDirectoryToSE(self):
    #   lfnDir = ['/this/is/dir1/', '/this/is/dir2/']
    #
    #   res = self.ci.catalogDirectoryToSE(lfnDir)
    #   self.assertTrue(res['OK'])

    def test__getCatalogDirectoryContents(self):
        lfnDirs = ["/this/is/dir1/", "/this/is/dir2/"]

        res = self.ci._getCatalogDirectoryContents(lfnDirs)
        self.assertTrue(res["OK"])

        resExpected = {
            "Metadata": {
                "/this/is/dir1/file1.txt": {
                    "MetaData": {
                        "Checksum": "7149ed85",
                        "ChecksumType": "Adler32",
                        "CreationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                        "FileID": 156301805,
                        "GID": 2695,
                        "GUID": "6A5C6C86-AD7B-E411-9EDB",
                        "Mode": 436,
                        "ModificationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                        "Owner": "phicharp",
                        "OwnerGroup": "lhcb_prod",
                        "Size": 206380531,
                        "Status": "AprioriGood",
                        "Type": "File",
                        "UID": 19503,
                    }
                },
                "/this/is/dir1/file2.foo.bar": {
                    "MetaData": {
                        "Checksum": "7149ed86",
                        "ChecksumType": "Adler32",
                        "CreationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                        "FileID": 156301805,
                        "GID": 2695,
                        "GUID": "6A5C6C86-AD7B-E411-9EDB",
                        "Mode": 436,
                        "ModificationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                        "Owner": "phicharp",
                        "OwnerGroup": "lhcb_prod",
                        "Size": 206380532,
                        "Status": "AprioriGood",
                        "Type": "File",
                        "UID": 19503,
                    }
                },
                "/this/is/dir2/subdir1/file3.pippo": {
                    "MetaData": {
                        "Checksum": "7149ed86",
                        "ChecksumType": "Adler32",
                        "CreationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                        "FileID": 156301805,
                        "GID": 2695,
                        "GUID": "6A5C6C86-AD7B-E411-9EDB",
                        "Mode": 436,
                        "ModificationDate": datetime.datetime(2014, 12, 4, 12, 16, 56),
                        "Owner": "phicharp",
                        "OwnerGroup": "lhcb_prod",
                        "Size": 206380532,
                        "Status": "AprioriGood",
                        "Type": "File",
                        "UID": 19503,
                    }
                },
            },
            "Replicas": {
                "/this/is/dir1/file1.txt": {
                    "SE1": "smr://srm.SE1.ch:8443/srm/v2/server?SFN=/this/is/dir1/file1.txt",
                    "SE2": "smr://srm.SE2.fr:8443/srm/v2/server?SFN=/this/is/dir1/file1.txt",
                },
                "/this/is/dir1/file2.foo.bar": {
                    "SE1": "smr://srm.SE1.ch:8443/srm/v2/server?SFN=/this/is/dir1/file2.foo.bar",
                    "SE3": "smr://srm.SE3.es:8443/srm/v2/server?SFN=/this/is/dir1/file2.foo.bar",
                },
            },
        }

        self.assertEqual(res["Value"], resExpected)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ConsistencyInspectorSuccess))
    testResult = unittest.TextTestRunner(verbosity=3).run(suite)



#!/usr/bin/env python
"""
Remove the outputs produced by a transformation
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["transID: transformation ID"])
    _, args = Script.parseCommandLine()

    transIDs = [int(arg) for arg in args]

    from DIRAC.TransformationSystem.Agent.TransformationCleaningAgent import TransformationCleaningAgent
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    agent = TransformationCleaningAgent(
        "Transformation/TransformationCleaningAgent",
        "Transformation/TransformationCleaningAgent",
        "dirac-transformation-remove-output",
    )
    agent.initialize()

    client = TransformationClient()
    for transID in transIDs:
        agent.removeTransformationOutput(transID)


if __name__ == "__main__":
    main()



""" The Machine/Job Features TimeLeft utility interrogates the MJF values
    for the current CPU and Wallclock consumed, as well as their limits.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import time
from six.moves.urllib.request import urlopen

from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.ResourceUsage import ResourceUsage


class MJFResourceUsage(ResourceUsage):
    """
    This is the MJF plugin of the TimeLeft Utility
    """

    #############################################################################
    def __init__(self):
        """Standard constructor"""
        super(MJFResourceUsage, self).__init__("MJF", "JOB_ID")

        self.queue = os.environ.get("QUEUE")

        self.log.verbose("jobID=%s, queue=%s" % (self.jobID, self.queue))
        self.startTime = time.time()

    #############################################################################
    def getResourceUsage(self):
        """Returns S_OK with a dictionary containing the entries CPU, CPULimit,
        WallClock, WallClockLimit, and Unit for current slot.
        """

        cpuLimit = None
        wallClockLimit = None
        wallClock = None
        jobStartSecs = None

        jobFeaturesPath = None
        machineFeaturesPath = None

        # Getting info from JOBFEATURES
        try:
            # We are not called from TimeLeft.py if these are not set
            jobFeaturesPath = os.environ["JOBFEATURES"]
        except KeyError:
            self.log.warn("$JOBFEATURES is not set")

        if jobFeaturesPath:
            try:
                wallClockLimit = int(urlopen(jobFeaturesPath + "/wall_limit_secs").read())
                self.log.verbose("wallClockLimit from JF = %d" % wallClockLimit)
            except ValueError:
                self.log.warn("/wall_limit_secs is unreadable")
            except IOError as e:
                self.log.exception("Issue with $JOBFEATURES/wall_limit_secs", lException=e)
                self.log.warn("Could not determine cpu limit from $JOBFEATURES/wall_limit_secs")

            try:
                jobStartSecs = int(urlopen(jobFeaturesPath + "/jobstart_secs").read())
                self.log.verbose("jobStartSecs from JF = %d" % jobStartSecs)
            except ValueError:
                self.log.warn("/jobstart_secs is unreadable, setting a default")
                jobStartSecs = self.startTime
            except IOError as e:
                self.log.exception("Issue with $JOBFEATURES/jobstart_secs", lException=e)
                self.log.warn("Can't open jobstart_secs, setting a default")
                jobStartSecs = self.startTime

            try:
                cpuLimit = int(urlopen(jobFeaturesPath + "/cpu_limit_secs").read())
                self.log.verbose("cpuLimit from JF = %d" % cpuLimit)
            except ValueError:
                self.log.warn("/cpu_limit_secs is unreadable")
            except IOError as e:
                self.log.exception("Issue with $JOBFEATURES/cpu_limit_secs", lException=e)
                self.log.warn("Could not determine cpu limit from $JOBFEATURES/cpu_limit_secs")

            wallClock = int(time.time()) - jobStartSecs

        # Getting info from MACHINEFEATURES
        try:
            # We are not called from TimeLeft.py if these are not set
            machineFeaturesPath = os.environ["MACHINEFEATURES"]
        except KeyError:
            self.log.warn("$MACHINEFEATURES is not set")

        if machineFeaturesPath and jobStartSecs:
            try:
                shutdownTime = int(urlopen(machineFeaturesPath + "/shutdowntime").read())
                self.log.verbose("shutdownTime from MF = %d" % shutdownTime)
                if int(time.time()) + wallClockLimit > shutdownTime:
                    # reduce wallClockLimit if would overrun shutdownTime
                    wallClockLimit = shutdownTime - jobStartSecs
            except ValueError:
                self.log.warn("/shutdowntime is unreadable")
            except IOError as e:
                self.log.warn("Issue with $MACHINEFEATURES/shutdowntime", repr(e))
                self.log.warn("Could not determine a shutdowntime value from $MACHINEFEATURES/shutdowntime")

        # Reporting
        consumed = {"CPU": None, "CPULimit": cpuLimit, "WallClock": wallClock, "WallClockLimit": wallClockLimit}
        if cpuLimit and wallClock and wallClockLimit:
            self.log.verbose("MJF consumed: %s" % str(consumed))
            return S_OK(consumed)
        self.log.info("Could not determine some parameters")
        retVal = S_ERROR("Could not determine some parameters")
        retVal["Value"] = consumed
        return retVal



"""
.. versionadded:: v6r20

FTS3Agent implementation.
It is in charge of submitting and monitoring all the transfers. It can be duplicated.


.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN FTS3Agent
  :end-before: ##END FTS3Agent
  :dedent: 2
  :caption: FTS3Agent options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import errno
import time

# from threading import current_thread
from multiprocessing.pool import ThreadPool

# We use the dummy module because we use the ThreadPool
from multiprocessing.dummy import current_process
from socket import gethostname

from DIRAC import S_OK, S_ERROR

from DIRAC.AccountingSystem.Client.Types.DataOperation import DataOperation
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.DErrno import cmpError
from DIRAC.Core.Utilities.DictCache import DictCache

from DIRAC.Core.Utilities.Time import fromString
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getFTS3ServerDict
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations as opHelper
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.DataManagementSystem.private import FTS3Utilities
from DIRAC.DataManagementSystem.DB.FTS3DB import FTS3DB
from DIRAC.DataManagementSystem.Client.FTS3Job import FTS3Job
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient


# pylint: disable=attribute-defined-outside-init

AGENT_NAME = "DataManagement/FTS3Agent"

# Lifetime in seconds of the proxy we download for submission
PROXY_LIFETIME = 43200  # 12 hours


class FTS3Agent(AgentModule):
    """
    This Agent is responsible of interacting with the FTS3 services.
    Several of them can run in parallel.
    It first treats the Operations, by creating new FTS jobs and performing
    callback.
    Then, it monitors the current jobs.

    CAUTION: This agent and the FTSAgent cannot run together.

    """

    def __readConf(self):
        """Read configurations

        :return: S_OK()/S_ERROR()
        """

        # Getting all the possible servers
        res = getFTS3ServerDict()
        if not res["OK"]:
            gLogger.error(res["Message"])
            return res

        srvDict = res["Value"]
        serverPolicyType = opHelper().getValue("DataManagement/FTSPlacement/FTS3/ServerPolicy", "Random")
        self._serverPolicy = FTS3Utilities.FTS3ServerPolicy(srvDict, serverPolicy=serverPolicyType)

        self.maxNumberOfThreads = self.am_getOption("MaxThreads", 10)

        # Number of Operation we treat in one loop
        self.operationBulkSize = self.am_getOption("OperationBulkSize", 20)
        # Number of Jobs we treat in one loop
        self.jobBulkSize = self.am_getOption("JobBulkSize", 20)
        self.maxFilesPerJob = self.am_getOption("MaxFilesPerJob", 100)
        self.maxAttemptsPerFile = self.am_getOption("MaxAttemptsPerFile", 256)
        self.kickDelay = self.am_getOption("KickAssignedHours", 1)
        self.maxKick = self.am_getOption("KickLimitPerCycle", 100)
        self.deleteDelay = self.am_getOption("DeleteGraceDays", 180)
        self.maxDelete = self.am_getOption("DeleteLimitPerCycle", 100)
        # lifetime of the proxy we download to delegate to FTS
        self.proxyLifetime = self.am_getOption("ProxyLifetime", PROXY_LIFETIME)

        return S_OK()

    def initialize(self):
        """Agent's initialization

        :return: S_OK()/S_ERROR()
        """

        self._globalContextCache = {}

        # name that will be used in DB for assignment tag
        self.assignmentTag = gethostname().split(".")[0]

        res = self.__readConf()

        # We multiply by two because of the two threadPools
        self.fts3db = FTS3DB(pool_size=2 * self.maxNumberOfThreads)

        self.jobsThreadPool = ThreadPool(self.maxNumberOfThreads)
        self.opsThreadPool = ThreadPool(self.maxNumberOfThreads)

        return res

    def beginExecution(self):
        """Reload configurations before start of a cycle

        :return: S_OK()/S_ERROR()
        """
        return self.__readConf()

    def getFTS3Context(self, username, group, ftsServer, threadID):
        """Returns an fts3 context for a given user, group and fts server

        The context pool is per thread, and there is one context
        per tuple (user, group, server).
        We dump the proxy of a user to a file (shared by all the threads),
        and use it to make the context.
        The proxy needs a lifetime of self.proxyLifetime, is cached for cacheTime = (2*lifeTime/3) - 10mn,
        and the lifetime of the context is 45mn
        The reason for cacheTime to be what it is is because the FTS3 server will ask for a new proxy
        after 2/3rd of the existing proxy has expired, so we renew it just before

        :param str username: name of the user
        :param str group: group of the user
        :param str ftsServer: address of the server
        :param str threadID: thread ID

        :returns: S_OK with the context object

        """

        log = gLogger.getSubLogger("getFTS3Context")

        contextes = self._globalContextCache.setdefault(threadID, DictCache())

        idTuple = (username, group, ftsServer)
        log.debug("Getting context for %s" % (idTuple,))

        # We keep a context in the cache for 45 minutes
        # (so it needs to be valid at least 15 since we add it for one hour)
        if not contextes.exists(idTuple, 15 * 60):
            res = getDNForUsername(username)
            if not res["OK"]:
                return res
            # We take the first DN returned
            userDN = res["Value"][0]

            log.debug("UserDN %s" % userDN)

            # We dump the proxy to a file.
            # It has to have a lifetime of self.proxyLifetime
            # Because the FTS3 servers cache it for 2/3rd of the lifetime
            # we should make our cache a bit less than 2/3rd of the lifetime
            cacheTime = int(2 * self.proxyLifetime / 3) - 600
            res = gProxyManager.downloadVOMSProxyToFile(
                userDN, group, requiredTimeLeft=self.proxyLifetime, cacheTime=cacheTime
            )
            if not res["OK"]:
                return res

            proxyFile = res["Value"]
            log.debug("Proxy file %s" % proxyFile)

            # We generate the context
            # In practice, the lifetime will be less than proxyLifetime
            # because we reuse a cached proxy. However, the cached proxy will
            # never forced a redelegation, because it is recent enough for FTS3 servers.
            # The delegation is forced when 2/3 rd of the lifetime are left, and we get a fresh
            # one just before. So no problem
            res = FTS3Job.generateContext(ftsServer, proxyFile, lifetime=self.proxyLifetime)

            if not res["OK"]:
                return res
            context = res["Value"]

            # we add it to the cache for this thread for 1h
            contextes.add(idTuple, 3600, context)

        return S_OK(contextes.get(idTuple))

    def _monitorJob(self, ftsJob):
        """* query the FTS servers
        * update the FTSFile status
        * update the FTSJob status

        :param ftsJob: FTS job

        :return: ftsJob, S_OK()/S_ERROR()
        """
        # General try catch to avoid that the tread dies
        try:
            threadID = current_process().name
            log = gLogger.getLocalSubLogger("_monitorJob/%s" % ftsJob.jobID)

            res = self.getFTS3Context(ftsJob.username, ftsJob.userGroup, ftsJob.ftsServer, threadID=threadID)

            if not res["OK"]:
                log.error("Error getting context", res)
                return ftsJob, res

            context = res["Value"]

            res = ftsJob.monitor(context=context)

            if not res["OK"]:
                log.error("Error monitoring job", res)

                # If the job was not found on the server, update the DB
                if cmpError(res, errno.ESRCH):
                    res = self.fts3db.cancelNonExistingJob(ftsJob.operationID, ftsJob.ftsGUID)

                return ftsJob, res

            # { fileID : { Status, Error } }
            filesStatus = res["Value"]

            # Specify the job ftsGUID to make sure we do not overwrite
            # status of files already taken by newer jobs
            res = self.fts3db.updateFileStatus(filesStatus, ftsGUID=ftsJob.ftsGUID)

            if not res["OK"]:
                log.error("Error updating file fts status", "%s, %s" % (ftsJob.ftsGUID, res))
                return ftsJob, res

            upDict = {
                ftsJob.jobID: {
                    "status": ftsJob.status,
                    "error": ftsJob.error,
                    "completeness": ftsJob.completeness,
                    "operationID": ftsJob.operationID,
                    "lastMonitor": True,
                }
            }
            res = self.fts3db.updateJobStatus(upDict)

            if ftsJob.status in ftsJob.FINAL_STATES:
                self.__sendAccounting(ftsJob)

            return ftsJob, res

        except Exception as e:
            log.exception("Exception while monitoring job", repr(e))
            return ftsJob, S_ERROR(0, "Exception %s" % repr(e))

    @staticmethod
    def _monitorJobCallback(returnedValue):
        """Callback when a job has been monitored
        :param returnedValue: value returned by the _monitorJob method
                              (ftsJob, standard dirac return struct)
        """

        ftsJob, res = returnedValue
        log = gLogger.getLocalSubLogger("_monitorJobCallback/%s" % ftsJob.jobID)
        if not res["OK"]:
            log.error("Error updating job status", res)
        else:
            log.debug("Successfully updated job status")

    def monitorJobsLoop(self):
        """* fetch the active FTSJobs from the DB
        * spawn a thread to monitor each of them

        :return: S_OK()/S_ERROR()
        """

        log = gLogger.getSubLogger("monitorJobs")
        log.debug("Size of the context cache %s" % len(self._globalContextCache))

        log.debug("Getting active jobs")
        # get jobs from DB
        res = self.fts3db.getActiveJobs(limit=self.jobBulkSize, jobAssignmentTag=self.assignmentTag)

        if not res["OK"]:
            log.error("Could not retrieve ftsJobs from the DB", res)
            return res

        activeJobs = res["Value"]
        log.info("%s jobs to queue for monitoring" % len(activeJobs))

        # We store here the AsyncResult object on which we are going to wait
        applyAsyncResults = []

        # Starting the monitoring threads
        for ftsJob in activeJobs:
            log.debug("Queuing executing of ftsJob %s" % ftsJob.jobID)
            # queue the execution of self._monitorJob( ftsJob ) in the thread pool
            # The returned value is passed to _monitorJobCallback
            applyAsyncResults.append(
                self.jobsThreadPool.apply_async(self._monitorJob, (ftsJob,), callback=self._monitorJobCallback)
            )

        log.debug("All execution queued")

        # Waiting for all the monitoring to finish
        while not all([r.ready() for r in applyAsyncResults]):
            log.debug("Not all the tasks are finished")
            time.sleep(0.5)

        log.debug("All the tasks have completed")
        return S_OK()

    @staticmethod
    def _treatOperationCallback(returnedValue):
        """Callback when an operation has been treated

        :param returnedValue: value returned by the _treatOperation method
                              (ftsOperation, standard dirac return struct)
        """

        operation, res = returnedValue
        log = gLogger.getLocalSubLogger("_treatOperationCallback/%s" % operation.operationID)
        if not res["OK"]:
            log.error("Error treating operation", res)
        else:
            log.debug("Successfully treated operation")

    def _treatOperation(self, operation):
        """Treat one operation:
          * does the callback if the operation is finished
          * generate new jobs and submits them

        :param operation: the operation to treat

        :return: operation, S_OK()/S_ERROR()
        """
        try:
            threadID = current_process().name
            log = gLogger.getLocalSubLogger("treatOperation/%s" % operation.operationID)

            # If the operation is totally processed
            # we perform the callback
            if operation.isTotallyProcessed():
                log.debug("FTS3Operation %s is totally processed" % operation.operationID)
                res = operation.callback()

                if not res["OK"]:
                    log.error("Error performing the callback", res)
                    log.info("Putting back the operation")
                    dbRes = self.fts3db.persistOperation(operation)

                    if not dbRes["OK"]:
                        log.error("Could not persist operation", dbRes)

                    return operation, res

            else:
                log.debug("FTS3Operation %s is not totally processed yet" % operation.operationID)

                # This flag is set to False if we want to stop the ongoing processing
                # of an operation, typically when the matching RMS Request has been
                # canceled (see below)
                continueOperationProcessing = True

                # Check the status of the associated RMS Request.
                # If it is canceled or does not exist anymore then we will not create new FTS3Jobs, and mark
                # this as FTS3Operation canceled.

                if operation.rmsReqID:
                    res = ReqClient().getRequestStatus(operation.rmsReqID)
                    if not res["OK"]:
                        # If the Request does not exist anymore
                        if cmpError(res, errno.ENOENT):
                            log.info(
                                "The RMS Request does not exist anymore, canceling the FTS3Operation",
                                "rmsReqID: %s, FTS3OperationID: %s" % (operation.rmsReqID, operation.operationID),
                            )
                            operation.status = "Canceled"
                            continueOperationProcessing = False
                        else:
                            log.error("Could not get request status", res)
                            return operation, res

                    else:
                        rmsReqStatus = res["Value"]

                        if rmsReqStatus == "Canceled":
                            log.info(
                                "The RMS Request is canceled, canceling the FTS3Operation",
                                "rmsReqID: %s, FTS3OperationID: %s" % (operation.rmsReqID, operation.operationID),
                            )
                            operation.status = "Canceled"
                            continueOperationProcessing = False

                if continueOperationProcessing:

                    res = operation.prepareNewJobs(
                        maxFilesPerJob=self.maxFilesPerJob, maxAttemptsPerFile=self.maxAttemptsPerFile
                    )

                    if not res["OK"]:
                        log.error("Cannot prepare new Jobs", "FTS3Operation %s : %s" % (operation.operationID, res))
                        return operation, res

                    newJobs = res["Value"]

                    log.debug("FTS3Operation %s: %s new jobs to be submitted" % (operation.operationID, len(newJobs)))

                    for ftsJob in newJobs:
                        res = self._serverPolicy.chooseFTS3Server()
                        if not res["OK"]:
                            log.error(res)
                            continue

                        ftsServer = res["Value"]
                        log.debug("Use %s server" % ftsServer)

                        ftsJob.ftsServer = ftsServer

                        res = self.getFTS3Context(ftsJob.username, ftsJob.userGroup, ftsServer, threadID=threadID)

                        if not res["OK"]:
                            log.error("Could not get context", res)
                            continue

                        context = res["Value"]

                        try:
                            tpcProtocols = operation.fts3Plugin.selectTPCProtocols(ftsJob=ftsJob)
                        except ValueError as e:
                            log.error("Could not select TPC list", repr(e))
                            continue

                        res = ftsJob.submit(context=context, protocols=tpcProtocols)

                        if not res["OK"]:
                            log.error(
                                "Could not submit FTS3Job", "FTS3Operation %s : %s" % (operation.operationID, res)
                            )
                            continue

                        operation.ftsJobs.append(ftsJob)

                        submittedFileIds = res["Value"]
                        log.info(
                            "FTS3Operation %s: Submitted job for %s transfers"
                            % (operation.operationID, len(submittedFileIds))
                        )

                # new jobs are put in the DB at the same time
            res = self.fts3db.persistOperation(operation)

            if not res["OK"]:
                log.error("Could not persist operation", res)

            return operation, res

        except Exception as e:
            log.exception("Exception in the thread", repr(e))
            return operation, S_ERROR("Exception %s" % repr(e))

    def treatOperationsLoop(self):
        """* Fetch all the FTSOperations which are not finished
        * Spawn a thread to treat each operation

        :return: S_OK()/S_ERROR()
        """

        log = gLogger.getSubLogger("treatOperations")

        log.debug("Size of the context cache %s" % len(self._globalContextCache))

        log.info("Getting non finished operations")

        res = self.fts3db.getNonFinishedOperations(
            limit=self.operationBulkSize, operationAssignmentTag=self.assignmentTag
        )

        if not res["OK"]:
            log.error("Could not get incomplete operations", res)
            return res

        incompleteOperations = res["Value"]

        log.info("Treating %s incomplete operations" % len(incompleteOperations))

        applyAsyncResults = []

        for operation in incompleteOperations:
            log.debug("Queuing executing of operation %s" % operation.operationID)
            # queue the execution of self._treatOperation( operation ) in the thread pool
            # The returned value is passed to _treatOperationCallback
            applyAsyncResults.append(
                self.opsThreadPool.apply_async(
                    self._treatOperation, (operation,), callback=self._treatOperationCallback
                )
            )

        log.debug("All execution queued")

        # Waiting for all the treatments to finish
        while not all([r.ready() for r in applyAsyncResults]):
            log.debug("Not all the tasks are finished")
            time.sleep(0.5)

        log.debug("All the tasks have completed")

        return S_OK()

    def kickOperations(self):
        """Kick stuck operations

        :return: S_OK()/S_ERROR()
        """

        log = gLogger.getSubLogger("kickOperations")

        res = self.fts3db.kickStuckOperations(limit=self.maxKick, kickDelay=self.kickDelay)
        if not res["OK"]:
            return res

        kickedOperations = res["Value"]
        log.info("Kicked %s stuck operations" % kickedOperations)

        return S_OK()

    def kickJobs(self):
        """Kick stuck jobs

        :return: S_OK()/S_ERROR()
        """

        log = gLogger.getSubLogger("kickJobs")

        res = self.fts3db.kickStuckJobs(limit=self.maxKick, kickDelay=self.kickDelay)
        if not res["OK"]:
            return res

        kickedJobs = res["Value"]
        log.info("Kicked %s stuck jobs" % kickedJobs)

        return S_OK()

    def deleteOperations(self):
        """Delete final operations

        :return: S_OK()/S_ERROR()
        """

        log = gLogger.getSubLogger("deleteOperations")

        res = self.fts3db.deleteFinalOperations(limit=self.maxDelete, deleteDelay=self.deleteDelay)
        if not res["OK"]:
            return res

        deletedOperations = res["Value"]
        log.info("Deleted %s final operations" % deletedOperations)

        return S_OK()

    def finalize(self):
        """Finalize processing

        :return: S_OK()/S_ERROR()
        """
        # Joining all the ThreadPools
        log = gLogger.getSubLogger("Finalize")

        log.debug("Closing jobsThreadPool")

        self.jobsThreadPool.close()
        self.jobsThreadPool.join()

        log.debug("jobsThreadPool joined")

        log.debug("Closing opsThreadPool")

        self.opsThreadPool.close()
        self.opsThreadPool.join()

        log.debug("opsThreadPool joined")

        return S_OK()

    def execute(self):
        """One cycle execution

        :return: S_OK()/S_ERROR()
        """

        log = gLogger.getSubLogger("execute")

        log.info("Monitoring job")
        res = self.monitorJobsLoop()

        if not res["OK"]:
            log.error("Error monitoring jobs", res)
            return res

        log.info("Treating operations")
        res = self.treatOperationsLoop()

        if not res["OK"]:
            log.error("Error treating operations", res)
            return res

        log.info("Kicking stuck jobs")
        res = self.kickJobs()

        if not res["OK"]:
            log.error("Error kicking jobs", res)
            return res

        log.info("Kicking stuck operations")
        res = self.kickOperations()

        if not res["OK"]:
            log.error("Error kicking operations", res)
            return res

        log.info("Deleting final operations")
        res = self.deleteOperations()

        if not res["OK"]:
            log.error("Error deleting operations", res)
            return res

        return S_OK()

    @staticmethod
    def __sendAccounting(ftsJob):
        """prepare and send DataOperation to AccountingDB

        :param ftsJob: the FTS3Job from which we send the accounting info
        """

        dataOp = DataOperation()
        dataOp.setStartTime(fromString(ftsJob.submitTime))
        dataOp.setEndTime(fromString(ftsJob.lastUpdate))

        dataOp.setValuesFromDict(ftsJob.accountingDict)
        dataOp.delayedCommit()



""" Module invoked for finding and loading DIRAC (and extensions) modules
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from DIRAC.Core.Utilities import List
from DIRAC import gConfig, S_ERROR, S_OK, gLogger
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Utilities.Extensions import extensionsByPriority, recurseImport


class ModuleLoader(object):
    def __init__(self, importLocation, sectionFinder, superClass, csSuffix=False, moduleSuffix=False):
        self.__modules = {}
        self.__loadedModules = {}
        self.__superClass = superClass
        # Function to find the
        self.__sectionFinder = sectionFinder
        # Import from where? <Ext>.<System>System.<importLocation>.<module>
        self.__importLocation = importLocation
        # Where to look in the CS for the module? /Systems/<System>/<Instance>/<csSuffix>
        if not csSuffix:
            csSuffix = "%ss" % importLocation
        self.__csSuffix = csSuffix
        # Module suffix (for Handlers)
        self.__modSuffix = moduleSuffix

    def getModules(self):
        data = dict(self.__modules)
        for k in data:
            data[k]["standalone"] = len(data) == 1
        return data

    def loadModules(self, modulesList, hideExceptions=False):
        """
        Load all modules required in moduleList
        """
        for modName in modulesList:
            gLogger.verbose("Checking %s" % modName)
            # if it's a executor modName name just load it and be done with it
            if "/" in modName:
                gLogger.verbose("Module %s seems to be a valid name. Try to load it!" % modName)
                result = self.loadModule(modName, hideExceptions=hideExceptions)
                if not result["OK"]:
                    return result
                continue
            # Check if it's a system name
            # Look in the CS
            system = modName
            # Can this be generated with sectionFinder?
            csPath = "%s/Executors" % PathFinder.getSystemSection(system)
            gLogger.verbose("Exploring %s to discover modules" % csPath)
            result = gConfig.getSections(csPath)
            if result["OK"]:
                # Add all modules in the CS :P
                for modName in result["Value"]:
                    result = self.loadModule("%s/%s" % (system, modName), hideExceptions=hideExceptions)
                    if not result["OK"]:
                        return result
            # Look what is installed
            parentModule = None
            for rootModule in extensionsByPriority():
                if not system.endswith("System"):
                    system += "System"
                parentImport = "%s.%s.%s" % (rootModule, system, self.__csSuffix)
                # HERE!
                result = recurseImport(parentImport)
                if not result["OK"]:
                    return result
                parentModule = result["Value"]
                if parentModule:
                    break
            if not parentModule:
                continue
            parentPath = parentModule.__path__[0]
            gLogger.notice("Found modules path at %s" % parentImport)
            for entry in os.listdir(parentPath):
                if entry == "__init__.py" or not entry.endswith(".py"):
                    continue
                if not os.path.isfile(os.path.join(parentPath, entry)):
                    continue
                modName = "%s/%s" % (system, entry[:-3])
                gLogger.verbose("Trying to import %s" % modName)
                result = self.loadModule(modName, hideExceptions=hideExceptions, parentModule=parentModule)

        return S_OK()

    def loadModule(self, modName, hideExceptions=False, parentModule=False):
        """
        Load module name.
        name must take the form [DIRAC System Name]/[DIRAC module]
        """
        while modName and modName[0] == "/":
            modName = modName[1:]
        if modName in self.__modules:
            return S_OK()
        modList = modName.split("/")
        if len(modList) != 2:
            return S_ERROR("Can't load %s: Invalid module name" % (modName))
        csSection = self.__sectionFinder(modName)
        loadGroup = gConfig.getValue("%s/Load" % csSection, [])
        # Check if it's a load group
        if loadGroup:
            gLogger.info("Found load group %s. Will load %s" % (modName, ", ".join(loadGroup)))
            for loadModName in loadGroup:
                if "/" not in loadModName:
                    loadModName = "%s/%s" % (modList[0], loadModName)
                result = self.loadModule(loadModName, hideExceptions=hideExceptions)
                if not result["OK"]:
                    return result
            return S_OK()
        # Normal load
        loadName = gConfig.getValue("%s/Module" % csSection, "")
        if not loadName:
            loadName = modName
            gLogger.info("Loading %s" % (modName))
        else:
            if "/" not in loadName:
                loadName = "%s/%s" % (modList[0], loadName)
            gLogger.info("Loading %s (%s)" % (modName, loadName))
        # If already loaded, skip
        loadList = loadName.split("/")
        if len(loadList) != 2:
            return S_ERROR("Can't load %s: Invalid module name" % (loadName))
        system, module = loadList
        # Load
        className = module
        if self.__modSuffix:
            className = "%s%s" % (className, self.__modSuffix)
        if loadName not in self.__loadedModules:
            # Check if handler is defined
            loadCSSection = self.__sectionFinder(loadName)
            handlerPath = gConfig.getValue("%s/HandlerPath" % loadCSSection, "")
            if handlerPath:
                gLogger.info("Trying to %s from CS defined path %s" % (loadName, handlerPath))
                gLogger.verbose("Found handler for %s: %s" % (loadName, handlerPath))
                handlerPath = handlerPath.replace("/", ".")
                if handlerPath.endswith(".py"):
                    handlerPath = handlerPath[:-3]
                className = List.fromChar(handlerPath, ".")[-1]
                result = recurseImport(handlerPath)
                if not result["OK"]:
                    return S_ERROR("Cannot load user defined handler %s: %s" % (handlerPath, result["Message"]))
                gLogger.verbose("Loading %s" % handlerPath)
            elif parentModule:
                gLogger.info("Trying to autodiscover %s from parent" % loadName)
                # If we've got a parent module, load from there.
                modImport = module
                if self.__modSuffix:
                    modImport = "%s%s" % (modImport, self.__modSuffix)
                result = recurseImport(modImport, parentModule, hideExceptions=hideExceptions)
            else:
                # Check to see if the module exists in any of the root modules
                gLogger.info("Trying to autodiscover %s" % loadName)
                for rootModule in extensionsByPriority():
                    importString = "%s.%sSystem.%s.%s" % (rootModule, system, self.__importLocation, module)
                    if self.__modSuffix:
                        importString = "%s%s" % (importString, self.__modSuffix)
                    gLogger.verbose("Trying to load %s" % importString)
                    result = recurseImport(importString, hideExceptions=hideExceptions)
                    # Error while loading
                    if not result["OK"]:
                        return result
                    # Something has been found! break :)
                    if result["Value"]:
                        gLogger.verbose("Found %s" % importString)
                        break
            # Nothing found
            if not result["Value"]:
                return S_ERROR("Could not find %s" % loadName)
            modObj = result["Value"]
            try:
                # Try to get the class from the module
                modClass = getattr(modObj, className)
            except AttributeError:
                if "__file__" in dir(modObj):
                    location = modObj.__file__
                else:
                    location = modObj.__path__
                gLogger.exception("%s module does not have a %s class!" % (location, module))
                return S_ERROR("Cannot load %s" % module)
            # Check if it's subclass
            if not issubclass(modClass, self.__superClass):
                return S_ERROR("%s has to inherit from %s" % (loadName, self.__superClass.__name__))
            self.__loadedModules[loadName] = {"classObj": modClass, "moduleObj": modObj}
            # End of loading of 'loadName' module

        # A-OK :)
        self.__modules[modName] = self.__loadedModules[loadName].copy()
        # keep the name of the real code module
        self.__modules[modName]["modName"] = modName
        self.__modules[modName]["loadName"] = loadName
        gLogger.notice("Loaded module %s" % modName)

        return S_OK()



""" This is a test of the DIRACCAProxyProvider
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=invalid-name,wrong-import-position,protected-access
import os
import re
import sys
import shutil

# TODO: This should be modernised to use subprocess(32)
try:
    import commands
except ImportError:
    # Python 3's subprocess module contains a compatibility layer
    import subprocess as commands
import unittest
import tempfile

import pytest

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Resources.ProxyProvider.DIRACCAProxyProvider import DIRACCAProxyProvider


certsPath = os.path.join(os.path.dirname(DIRAC.__file__), "Core/Security/test/certs")

testCAPath = os.path.join(tempfile.mkdtemp(dir="/tmp"), "ca")
testCAConfigFile = os.path.join(testCAPath, "openssl_config_ca.cnf")

diracCADict = {
    "ProviderType": "DIRACCA",
    "CertFile": os.path.join(certsPath, "ca/ca.cert.pem"),
    "KeyFile": os.path.join(certsPath, "ca/ca.key.pem"),
    "Supplied": ["O", "OU", "CN"],
    "Optional": ["emailAddress"],
    "DNOrder": ["O", "OU", "CN", "emailAddress"],
    "OU": "CA",
    "C": "DN",
    "O": "DIRACCA",
    "ProviderName": "DIRAC_CA",
}
diracCAConf = {"ProviderType": "DIRACCA", "CAConfigFile": testCAConfigFile, "ProviderName": "DIRAC_CA_CFG"}


class DIRACCAProviderTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.failed = False

        shutil.copytree(os.path.join(certsPath, "ca"), testCAPath)

        # Parse
        lines = []
        with open(testCAConfigFile, "r") as caCFG:
            for line in caCFG:
                if re.findall("=", re.sub(r"#.*", "", line)):
                    # Ignore comments
                    field = re.sub(r"#.*", "", line).replace(" ", "").rstrip().split("=")[0]
                    # Put the right dir
                    line = "dir = %s #PUT THE RIGHT DIR HERE!\n" % (testCAPath) if field == "dir" else line
                lines.append(line)
        # Write modified conf. file
        with open(testCAConfigFile, "w") as caCFG:
            caCFG.writelines(lines)

        # Result
        status, output = commands.getstatusoutput("ls -al %s" % testCAPath)
        if status:
            gLogger.error(output)
            exit()
        gLogger.debug("Test path:\n", output)

    def setUp(self):
        gLogger.debug("\n")
        if self.failed:
            self.fail(self.failed)

    def tearDown(self):
        pass

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(testCAPath):
            shutil.rmtree(testCAPath)


class testDIRACCAProvider(DIRACCAProviderTestCase):
    @pytest.mark.slow
    def test_getProxy(self):
        """Test 'getProxy' - try to get proxies for different users and check it"""

        def check(proxyStr, proxyProvider, name):
            """Check proxy

            :param str proxyStr: proxy as string
            :param str proxyProvider: proxy provider name
            :param str name: proxy name
            """
            proxyFile = os.path.join(testCAPath, proxyProvider + name.replace(" ", "") + ".pem")
            gLogger.info("Check proxy..")
            chain = X509Chain()
            result = chain.loadProxyFromString(proxyStr)
            self.assertTrue(result["OK"], "\n" + result.get("Message", "Error message is absent."))
            for result in [
                chain.getRemainingSecs(),
                chain.getIssuerCert(),
                chain.getPKeyObj(),
                chain.getCertList(),
                chain.getNumCertsInChain(),
                chain.generateProxyToString(3600),
                chain.generateProxyToFile(proxyFile, 3600),
                chain.isProxy(),
                chain.isLimitedProxy(),
                chain.isValidProxy(),
                chain.isVOMS(),
                chain.isRFC(),
            ]:
                self.assertTrue(result["OK"], "\n" + result.get("Message", "Error message is absent."))

        for proxyProvider, log in [
            (diracCADict, "configuring only in DIRAC CFG"),
            (diracCAConf, "read configuration file"),
        ]:
            gLogger.info("\n* Try proxy provider that %s.." % log)
            ca = DIRACCAProxyProvider()
            result = ca.setParameters(proxyProvider)
            self.assertTrue(result["OK"], "\n" + result.get("Message", "Error message is absent."))

            gLogger.info("* Get proxy using FullName and Email of user..")
            for name, email, res in [
                ("MrUser", "good@mail.com", True),
                ("MrUser_1", "good_1@mail.com", True),
                (False, "good@mail.com", False),
                ("MrUser", False, True),
            ]:
                gLogger.info("\nFullName: %s" % name or "absent", "Email: %s.." % email or "absent")
                # Create user DN
                result = ca.generateDN(FullName=name, Email=email)
                text = "Must be ended %s%s" % (
                    "successful" if res else "with error",
                    ": %s" % result.get("Message", "Error message is absent."),
                )
                self.assertEqual(result["OK"], res, text)
                if not res:
                    gLogger.info("Msg: %s" % (result["Message"]))
                else:
                    userDN = result["Value"]
                    gLogger.info("Created DN:", userDN)

                    result = ca.getProxy(userDN)
                    text = "Must be ended %s%s" % (
                        "successful" if res else "with error",
                        ": %s" % result.get("Message", "Error message is absent."),
                    )
                    self.assertEqual(result["OK"], res, text)
                    if not res:
                        gLogger.info("Msg: %s" % (result["Message"]))
                    else:
                        check(result["Value"], proxyProvider["ProviderName"], name)

            gLogger.info("\n* Get proxy using user DN..")
            for dn, name, res in [
                ("/O=DIRAC/OU=DIRAC CA/CN=user_3/emailAddress=some@mail.org", "user_3", True),
                ("/O=Dirac/OU=DIRAC CA/CN=user/emailAddress=some@mail.org", "user", True),
                ("/O=Dirac/OU=Without supplied field/emailAddress=some@mail.org", "not_suplied", False),
                ("/O=Dirac/OU=DIRAC CA/CN=without email", "no_email", True),
                ("/some=bad/DN=", "badDN", False),
                ("/BF=Bad Field/O=IN/CN=DN", "badField", False),
                (False, "absent", False),
            ]:
                gLogger.info("\nDN:", dn or "absent")
                try:
                    result = ca.getProxy(dn)
                except Exception as e:
                    result["Message"] = str(e)
                    self.assertFalse(res, e)
                text = "Must be ended %s%s" % (
                    "successful" if res else "with error",
                    ": %s" % result.get("Message", "Error message is absent."),
                )
                self.assertEqual(result["OK"], res, text)
                if not res:
                    gLogger.info("Msg: %s" % (result["Message"]))
                else:
                    check(result["Value"], proxyProvider["ProviderName"], name)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(DIRACCAProviderTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(testDIRACCAProvider))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)
    sys.exit(not testResult.wasSuccessful())



#!/usr/bin/env python
import fnmatch
import os
from pathlib import Path
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Optional

import click
import git
import typer
import yaml
from packaging.version import Version
from typer import colors as c

# Editable configuration
DEFAULT_HOST_OS = "cc7"
DEFAULT_MYSQL_VER = "mysql:8.0"
DEFAULT_ES_VER = "elasticsearch:7.9.1"
FEATURE_VARIABLES = {
    "DIRACOSVER": "master",
    "DIRACOS_TARBALL_PATH": None,
    "TEST_HTTPS": "No",
    "DIRAC_FEWER_CFG_LOCKS": None,
    "USE_PYTHON3": None,
}
DEFAULT_MODULES = {
    "DIRAC": Path(__file__).parent.absolute(),
}

# Static configuration
DB_USER = "Dirac"
DB_PASSWORD = "Dirac"
DB_ROOTUSER = "root"
DB_ROOTPWD = "password"
DB_HOST = "mysql"
DB_PORT = "3306"

# Implementation details
LOG_LEVEL_MAP = {
    "ALWAYS": (c.BLACK, c.WHITE),
    "NOTICE": (None, c.MAGENTA),
    "INFO": (None, c.GREEN),
    "VERBOSE": (None, c.CYAN),
    "DEBUG": (None, c.BLUE),
    "WARN": (None, c.YELLOW),
    "ERROR": (None, c.RED),
    "FATAL": (c.RED, c.BLACK),
}
LOG_PATTERN = re.compile(r"^[\d\-]{10} [\d:]{8} UTC [^\s]+ ([A-Z]+):")


class NaturalOrderGroup(click.Group):
    """Group for showing subcommands in the correct order"""

    def list_commands(self, ctx):
        return self.commands.keys()


app = typer.Typer(
    cls=NaturalOrderGroup,
    help=f"""Run the DIRAC integration tests.

A local DIRAC setup can be created and tested by running:

\b
  ./integration_tests.py create

This is equivalent to running:

\b
  ./integration_tests.py prepare-environment
  ./integration_tests.py install-server
  ./integration_tests.py install-client
  ./integration_tests.py test-server
  ./integration_tests.py test-client

The test setup can be shutdown using:

\b
  ./integration_tests.py destroy

See below for additional subcommands which are useful during local development.

## Features

The currently known features and their default values are:

\b
  HOST_OS: {DEFAULT_HOST_OS!r}
  MYSQL_VER: {DEFAULT_MYSQL_VER!r}
  ES_VER: {DEFAULT_ES_VER!r}
  {(os.linesep + '  ').join(['%s: %r' % x for x in FEATURE_VARIABLES.items()])}

All features can be prefixed with "SERVER_" or "CLIENT_" to limit their scope.

## Extensions

Integration tests can be ran for extensions to DIRAC by specifying the module
name and path such as:

\b
  ./integration_tests.py create --extra-module MyDIRAC=/path/to/MyDIRAC

This will modify the setup process based on the contents of
`MyDIRAC/tests/.dirac-ci-config.yaml`. See the Vanilla DIRAC file for the
available options.

## Command completion

Command completion of typer based scripts can be enabled by running:

  typer --install-completion

After restarting your terminal you command completion is available using:

  typer ./integration_tests.py run ...
""",
)


@app.command()
def create(
    flags: Optional[list[str]] = typer.Argument(None),
    editable: Optional[bool] = None,
    extra_module: Optional[list[str]] = None,
    release_var: Optional[str] = None,
    run_server_tests: bool = True,
    run_client_tests: bool = True,
):
    """Start a local instance of the integration tests"""
    prepare_environment(flags, editable, extra_module, release_var)
    install_server()
    install_client()
    exit_code = 0
    if run_server_tests:
        try:
            test_server()
        except TestExit as e:
            exit_code += e.exit_code
        else:
            raise NotImplementedError()
    if run_client_tests:
        try:
            test_client()
        except TestExit as e:
            exit_code += e.exit_code
        else:
            raise NotImplementedError()
    if exit_code != 0:
        typer.secho("One or more tests failed", err=True, fg=c.RED)
    raise typer.Exit(exit_code)


@app.command()
def destroy():
    """Destroy a local instance of the integration tests"""
    typer.secho("Shutting down and removing containers", err=True, fg=c.GREEN)
    with _gen_docker_compose(DEFAULT_MODULES) as docker_compose_fn:
        os.execvpe(
            "docker-compose",
            ["docker-compose", "-f", docker_compose_fn, "down", "--remove-orphans", "-t", "0"],
            _make_env({}),
        )


@app.command()
def prepare_environment(
    flags: Optional[list[str]] = typer.Argument(None),
    editable: Optional[bool] = None,
    extra_module: Optional[list[str]] = None,
    release_var: Optional[str] = None,
):
    """Prepare the local environment for installing DIRAC."""

    _check_containers_running(is_up=False)
    if editable is None:
        editable = sys.stdout.isatty()
        typer.secho(
            f"No value passed for --[no-]editable, automatically detected: {editable}",
            fg=c.YELLOW,
        )
    typer.echo(f"Preparing environment")

    modules = DEFAULT_MODULES | dict(f.split("=", 1) for f in extra_module)
    modules = {k: Path(v).absolute() for k, v in modules.items()}

    flags = dict(f.split("=", 1) for f in flags)
    docker_compose_env = _make_env(flags)
    server_flags = {}
    client_flags = {}
    for key, value in flags.items():
        if key.startswith("SERVER_"):
            server_flags[key[len("SERVER_") :]] = value
        elif key.startswith("CLIENT_"):
            client_flags[key[len("CLIENT_") :]] = value
        else:
            server_flags[key] = value
            client_flags[key] = value
    server_config = _make_config(modules, server_flags, release_var, editable)
    client_config = _make_config(modules, client_flags, release_var, editable)

    typer.secho("Running docker-compose to create containers", fg=c.GREEN)
    with _gen_docker_compose(modules) as docker_compose_fn:
        subprocess.run(
            ["docker-compose", "-f", docker_compose_fn, "up", "-d"],
            check=True,
            env=docker_compose_env,
        )

    typer.secho("Creating users in server and client containers", fg=c.GREEN)
    for container_name in ["server", "client"]:
        if os.getuid() == 0:
            continue
        cmd = _build_docker_cmd(container_name, use_root=True, cwd="/")
        gid = str(os.getgid())
        uid = str(os.getuid())
        ret = subprocess.run(cmd + ["groupadd", "--gid", gid, "dirac"], check=False)
        if ret.returncode != 0:
            typer.secho(f"Failed to add add group dirac with id={gid}", fg=c.YELLOW)
        subprocess.run(
            cmd
            + [
                "useradd",
                "--uid",
                uid,
                "--gid",
                gid,
                "-s",
                "/bin/bash",
                "-d",
                "/home/dirac",
                "dirac",
            ],
            check=True,
        )
        subprocess.run(cmd + ["chown", "dirac", "/home/dirac"], check=True)

    typer.secho("Creating MySQL user", fg=c.GREEN)
    cmd = ["docker", "exec", "mysql", "mysql", f"--password={DB_ROOTPWD}", "-e"]
    # It sometimes takes a while for MySQL to be ready so wait for a while if needed
    for _ in range(10):
        ret = subprocess.run(
            cmd + [f"CREATE USER '{DB_USER}'@'%' IDENTIFIED BY '{DB_PASSWORD}';"],
            check=False,
        )
        if ret.returncode != 0:
            typer.secho("Failed to connect to MySQL, will retry in 10 seconds", fg=c.YELLOW)
            time.sleep(10)
        break
    else:
        raise Exception(ret)
    subprocess.run(
        cmd + [f"CREATE USER '{DB_USER}'@'localhost' IDENTIFIED BY '{DB_PASSWORD}';"],
        check=True,
    )
    subprocess.run(
        cmd + [f"CREATE USER '{DB_USER}'@'mysql' IDENTIFIED BY '{DB_PASSWORD}';"],
        check=True,
    )

    typer.secho("Copying files to containers", fg=c.GREEN)
    for name, config in [("server", server_config), ("client", client_config)]:
        if path := config.get("DIRACOS_TARBALL_PATH"):
            path = Path(path)
            if config["USE_PYTHON3"]:
                config["DIRACOS_TARBALL_PATH"] = f"/{path.name}"
                subprocess.run(
                    ["docker", "cp", str(path), f"{name}:/{config['DIRACOS_TARBALL_PATH']}"],
                    check=True,
                )
            else:
                md5_fn = Path(str(path).replace(".tar.gz", ".md5"))
                if not md5_fn.exists():
                    typer.secho(
                        "Failed to find MD5 filename for DIRACOS_TARBALL_PATH. " f"Expected at: {md5_fn}",
                        err=True,
                        fg=c.RED,
                    )
                    raise typer.Exit(1)
                subprocess.run(["docker", "cp", str(path), f"{name}:/{path.name}"], check=True)
                subprocess.run(["docker", "cp", str(md5_fn), f"{name}:/{md5_fn.name}"], check=True)
                config["DIRACOS_TARBALL_PATH"] = "/"
                config["DIRACOSVER"] = md5_fn.stem.split("-", 1)[1]

        config_as_shell = _dict_to_shell(config)
        typer.secho(f"## {name.title()} config is:", fg=c.BRIGHT_WHITE, bg=c.BLACK)
        typer.secho(config_as_shell)

        with tempfile.TemporaryDirectory() as tmpdir:
            path = Path(tmpdir) / "CONFIG"
            path.write_text(config_as_shell)
            subprocess.run(
                ["docker", "cp", str(path), f"{name}:/home/dirac"],
                check=True,
            )

    for module_name, module_configs in _load_module_configs(modules).items():
        for command in module_configs.get("commands", {}).get("post-prepare", []):
            typer.secho(
                f"Running post-prepare command for {module_name}: {command}",
                err=True,
                fg=c.GREEN,
            )
            subprocess.run(command, check=True, shell=True)


@app.command()
def install_server():
    """Install DIRAC in the server container."""
    _check_containers_running()

    typer.secho("Running server installation", fg=c.GREEN)
    base_cmd = _build_docker_cmd("server", tty=False)
    subprocess.run(
        base_cmd + ["bash", "/home/dirac/LocalRepo/TestCode/DIRAC/tests/CI/install_server.sh"],
        check=True,
    )

    typer.secho("Copying credentials and certificates", fg=c.GREEN)
    base_cmd = _build_docker_cmd("client", tty=False)
    subprocess.run(
        base_cmd
        + [
            "mkdir",
            "-p",
            "/home/dirac/ServerInstallDIR/user",
            "/home/dirac/ClientInstallDIR/etc",
            "/home/dirac/.globus",
        ],
        check=True,
    )
    for path in [
        "etc/grid-security",
        "user/client.pem",
        "user/client.key",
        f"/tmp/x509up_u{os.getuid()}",
    ]:
        source = os.path.join("/home/dirac/ServerInstallDIR", path)
        ret = subprocess.run(
            ["docker", "cp", f"server:{source}", "-"],
            check=True,
            text=False,
            stdout=subprocess.PIPE,
        )
        if path.startswith("user/"):
            dest = f"client:/home/dirac/ServerInstallDIR/{os.path.dirname(path)}"
        elif path.startswith("/"):
            dest = f"client:{os.path.dirname(path)}"
        else:
            dest = f"client:/home/dirac/ClientInstallDIR/{os.path.dirname(path)}"
        subprocess.run(["docker", "cp", "-", dest], check=True, text=False, input=ret.stdout)
    subprocess.run(
        base_cmd
        + [
            "bash",
            "-c",
            "cp /home/dirac/ServerInstallDIR/user/client.* /home/dirac/.globus/",
        ],
        check=True,
    )


@app.command()
def install_client():
    """Install DIRAC in the client container."""
    _check_containers_running()
    typer.secho("Running client installation", fg=c.GREEN)
    base_cmd = _build_docker_cmd("client")
    subprocess.run(
        base_cmd + ["bash", "/home/dirac/LocalRepo/TestCode/DIRAC/tests/CI/install_client.sh"],
        check=True,
    )


@app.command()
def test_server():
    """Run the server integration tests."""
    _check_containers_running()
    typer.secho("Running server tests", err=True, fg=c.GREEN)
    base_cmd = _build_docker_cmd("server")
    ret = subprocess.run(base_cmd + ["bash", "TestCode/DIRAC/tests/CI/run_tests.sh"], check=False)
    color = c.GREEN if ret.returncode == 0 else c.RED
    typer.secho(f"Server tests finished with {ret.returncode}", err=True, fg=color)
    raise TestExit(ret.returncode)


@app.command()
def test_client():
    """Run the client integration tests."""
    _check_containers_running()
    typer.secho("Running client tests", err=True, fg=c.GREEN)
    base_cmd = _build_docker_cmd("client")
    ret = subprocess.run(base_cmd + ["bash", "TestCode/DIRAC/tests/CI/run_tests.sh"], check=False)
    color = c.GREEN if ret.returncode == 0 else c.RED
    typer.secho(f"Client tests finished with {ret.returncode}", err=True, fg=color)
    raise TestExit(ret.returncode)


@app.command()
def exec_server():
    """Start an interactive session in the server container."""
    _check_containers_running()
    cmd = _build_docker_cmd("server")
    cmd += [
        "bash",
        "-c",
        ". $HOME/CONFIG && . $HOME/ServerInstallDIR/bashrc && exec bash",
    ]
    typer.secho("Opening prompt inside server container", err=True, fg=c.GREEN)
    os.execvp(cmd[0], cmd)


@app.command()
def exec_client():
    """Start an interactive session in the client container."""
    _check_containers_running()
    cmd = _build_docker_cmd("client")
    cmd += [
        "bash",
        "-c",
        ". $HOME/CONFIG && . $HOME/ClientInstallDIR/bashrc && exec bash",
    ]
    typer.secho("Opening prompt inside client container", err=True, fg=c.GREEN)
    os.execvp(cmd[0], cmd)


@app.command()
def exec_mysql():
    """Start an interactive session in the server container."""
    _check_containers_running()
    cmd = _build_docker_cmd("mysql", use_root=True, cwd="/")
    cmd += [
        "bash",
        "-c",
        f"exec mysql --user={DB_USER} --password={DB_PASSWORD}",
    ]
    typer.secho("Opening prompt inside server container", err=True, fg=c.GREEN)
    os.execvp(cmd[0], cmd)


@app.command()
def list_services():
    """List the services which have been running.

    Only the services for which /log/current exists are shown.
    """
    _check_containers_running()
    typer.secho("Known services:", err=True)
    for service in _list_services()[1]:
        typer.secho(f"* {service}", err=True)


@app.command()
def runsvctrl(command: str, pattern: str):
    """Execute runsvctrl inside the server container."""
    _check_containers_running()
    runit_dir, services = _list_services()
    cmd = _build_docker_cmd("server", cwd=runit_dir)
    services = fnmatch.filter(services, pattern)
    if not services:
        typer.secho(f"No services match {pattern!r}", fg=c.RED)
        raise typer.Exit(code=1)
    cmd += ["runsvctrl", command] + services
    os.execvp(cmd[0], cmd)


@app.command()
def logs(pattern: str = "*", lines: int = 10, follow: bool = True):
    """Show DIRAC's logs from the service container.

    For services matching [--pattern] show the most recent [--lines] from the
    logs. If [--follow] is True, continiously stream the logs.
    """
    _check_containers_running()
    runit_dir, services = _list_services()
    base_cmd = _build_docker_cmd("server", tty=False) + ["tail"]
    base_cmd += [f"--lines={lines}"]
    if follow:
        base_cmd += ["-f"]
    with ThreadPoolExecutor(len(services)) as pool:
        for service in fnmatch.filter(services, pattern):
            cmd = base_cmd + [f"{runit_dir}/{service}/log/current"]
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=None, text=True)
            pool.submit(_log_popen_stdout, p)


class TestExit(typer.Exit):
    pass


@contextmanager
def _gen_docker_compose(modules):
    # Load the docker-compose configuration and mount the necessary volumes
    input_fn = Path(__file__).parent / "tests/CI/docker-compose.yml"
    docker_compose = yaml.safe_load(input_fn.read_text())
    volumes = [f"{path}:/home/dirac/LocalRepo/ALTERNATIVE_MODULES/{name}" for name, path in modules.items()]
    volumes += [f"{path}:/home/dirac/LocalRepo/TestCode/{name}" for name, path in modules.items()]
    docker_compose["services"]["dirac-server"]["volumes"] = volumes[:]
    docker_compose["services"]["dirac-client"]["volumes"] = volumes[:]

    # Add any extension services
    for module_name, module_configs in _load_module_configs(modules).items():
        for service_name, service_config in module_configs["extra-services"].items():
            typer.secho(f"Adding service {service_name} for {module_name}", err=True, fg=c.GREEN)
            docker_compose["services"][service_name] = service_config.copy()
            docker_compose["services"][service_name]["volumes"] = volumes[:]

    # Write to a tempory file with the appropriate profile name
    prefix = "ci"
    with tempfile.TemporaryDirectory() as tmpdir:
        input_docker_compose_dir = Path(__file__).parent / "tests/CI/"
        output_fn = Path(tmpdir) / prefix / "docker-compose.yml"
        output_fn.parent.mkdir()
        output_fn.write_text(yaml.safe_dump(docker_compose, sort_keys=False))
        shutil.copytree(input_docker_compose_dir / "envs", str(Path(tmpdir) / prefix), dirs_exist_ok=True)
        yield output_fn


def _check_containers_running(*, is_up=True):
    with _gen_docker_compose(DEFAULT_MODULES) as docker_compose_fn:
        running_containers = subprocess.run(
            ["docker-compose", "-f", docker_compose_fn, "ps", "-q", "-a"],
            stdout=subprocess.PIPE,
            env=_make_env({}),
            check=True,
            text=True,
        ).stdout.split("\n")
    if is_up:
        if not any(running_containers):
            typer.secho(
                f"No running containers found, environment must be prepared first!",
                err=True,
                fg=c.RED,
            )
            raise typer.Exit(code=1)
    else:
        if any(running_containers):
            typer.secho(
                f"Running instance already found, it must be destroyed first!",
                err=True,
                fg=c.RED,
            )
            raise typer.Exit(code=1)


def _find_dirac_release_and_branch():
    # Start by looking for the GitHub/GitLab environment variables
    ref = os.environ.get("CI_COMMIT_REF_NAME", os.environ.get("GITHUB_REF"))
    if ref == "refs/heads/integration":
        return "integration", ""
    ref = os.environ.get("CI_MERGE_REQUEST_TARGET_BRANCH_NAME", os.environ.get("GITHUB_BASE_REF"))
    if ref == "integration":
        return "integration", ""

    repo = git.Repo(os.getcwd())
    # Try to make sure the upstream remote is up to date
    try:
        upstream = repo.remote("upstream")
    except ValueError:
        typer.secho("No upstream remote found, adding", err=True, fg=c.YELLOW)
        upstream = repo.create_remote("upstream", "https://github.com/DIRACGrid/DIRAC.git")
    try:
        upstream.fetch()
    except Exception:
        typer.secho("Failed to fetch from remote 'upstream'", err=True, fg=c.YELLOW)
    # Find the most recent tag on the current branch
    version = Version(
        repo.git.describe(
            dirty=True,
            tags=True,
            long=True,
            match="*[0-9]*",
            exclude=["v[0-9]r*", "v[0-9][0-9]r*"],
        ).split("-")[0]
    )
    # See if there is a remote branch named "rel-vXrY"
    version_branch = f"rel-v{version.major}r{version.minor}"
    try:
        upstream.refs[version_branch]
    except IndexError:
        typer.secho(
            f"Failed to find branch for {version_branch}, defaulting to integration",
            err=True,
            fg=c.YELLOW,
        )
        return "integration", ""
    else:
        return "", f"v{version.major}r{version.minor}"


def _make_env(flags):
    env = os.environ.copy()
    env["DIRAC_UID"] = str(os.getuid())
    env["DIRAC_GID"] = str(os.getgid())
    env["HOST_OS"] = flags.pop("HOST_OS", DEFAULT_HOST_OS)
    env["CI_REGISTRY_IMAGE"] = flags.pop("CI_REGISTRY_IMAGE", "diracgrid")
    env["MYSQL_VER"] = flags.pop("MYSQL_VER", DEFAULT_MYSQL_VER)
    env["ES_VER"] = flags.pop("ES_VER", DEFAULT_ES_VER)
    return env


def _dict_to_shell(variables):
    lines = []
    for name, value in variables.items():
        if value is None:
            continue
        elif isinstance(value, list):
            lines += [f"declare -a {name}"]
            lines += [f"{name}+=({shlex.quote(v)})" for v in value]
        elif isinstance(value, bool):
            lines += [f"export {name}={'Yes' if value else 'No'}"]
        elif isinstance(value, str):
            lines += [f"export {name}={shlex.quote(value)}"]
        else:
            raise NotImplementedError(name, value, type(value))
    return "\n".join(lines)


def _make_config(modules, flags, release_var, editable):
    config = {
        "DEBUG": "True",
        # MYSQL Settings
        "DB_USER": DB_USER,
        "DB_PASSWORD": DB_PASSWORD,
        "DB_ROOTUSER": DB_ROOTUSER,
        "DB_ROOTPWD": DB_ROOTPWD,
        "DB_HOST": DB_HOST,
        "DB_PORT": DB_PORT,
        # ElasticSearch settings
        "NoSQLDB_HOST": "elasticsearch",
        "NoSQLDB_PORT": "9200",
        # Hostnames
        "SERVER_HOST": "server",
        "CLIENT_HOST": "client",
        # Test specific variables
        "WORKSPACE": "/home/dirac",
    }

    if editable:
        config["PIP_INSTALL_EXTRA_ARGS"] = "-e"

    required_feature_flags = []
    for module_name, module_ci_config in _load_module_configs(modules).items():
        config |= module_ci_config["config"]
        required_feature_flags += module_ci_config.get("required-feature-flags", [])
    config["DIRAC_CI_SETUP_SCRIPT"] = "/home/dirac/LocalRepo/TestCode/" + config["DIRAC_CI_SETUP_SCRIPT"]

    # This can likely be removed after the Python 3 migration
    if release_var:
        config |= dict([release_var.split("=", 1)])
    else:
        config["DIRAC_RELEASE"], config["DIRACBRANCH"] = _find_dirac_release_and_branch()

    for key, default_value in FEATURE_VARIABLES.items():
        config[key] = flags.pop(key, default_value)
    for key in required_feature_flags:
        try:
            config[key] = flags.pop(key)
        except KeyError:
            typer.secho(f"Required feature variable {key!r} is missing", err=True, fg=c.RED)
            raise typer.Exit(code=1)
    config["TESTREPO"] = [f"/home/dirac/LocalRepo/TestCode/{name}" for name in modules]
    config["ALTERNATIVE_MODULES"] = [f"/home/dirac/LocalRepo/ALTERNATIVE_MODULES/{name}" for name in modules]
    if not config["USE_PYTHON3"]:
        config["ALTERNATIVE_MODULES"] = [f"{x}/src/{Path(x).name}" for x in config["ALTERNATIVE_MODULES"]]

    # Exit with an error if there are unused feature flags remaining
    if flags:
        typer.secho(f"Unrecognised feature flags {flags!r}", err=True, fg=c.RED)
        raise typer.Exit(code=1)

    return config


def _load_module_configs(modules):
    module_ci_configs = {}
    for module_name, module_path in modules.items():
        module_ci_config_path = module_path / "tests/.dirac-ci-config.yaml"
        if not module_ci_config_path.exists():
            continue
        module_ci_configs[module_name] = yaml.safe_load(module_ci_config_path.read_text())
    return module_ci_configs


def _build_docker_cmd(container_name, *, use_root=False, cwd="/home/dirac", tty=True):
    if use_root or os.getuid() == 0:
        user = "root"
    else:
        user = "dirac"
    cmd = ["docker", "exec"]
    if tty:
        if sys.stdout.isatty():
            cmd += ["-it"]
        else:
            typer.secho(
                'Not passing "-it" to docker as stdout is not a tty',
                err=True,
                fg=c.YELLOW,
            )
    cmd += [
        "-e=TERM=xterm-color",
        "-e=INSTALLROOT=/home/dirac",
        f"-e=INSTALLTYPE={container_name}",
        f"-u={user}",
        f"-w={cwd}",
        container_name,
    ]
    return cmd


def _list_services():
    # The Python 3 runit dir ends up in /diracos
    for runit_dir in ["ServerInstallDIR/runit", "ServerInstallDIR/diracos/runit"]:
        cmd = _build_docker_cmd("server")
        cmd += [
            "bash",
            "-c",
            f'cd {runit_dir}/ && for fn in */*/log/current; do echo "$(dirname "$(dirname "$fn")")"; done',
        ]
        ret = subprocess.run(cmd, check=False, stdout=subprocess.PIPE, text=True)
        if not ret.returncode:
            return runit_dir, ret.stdout.split()
    else:
        typer.secho("Failed to find list of available services", err=True, fg=c.RED)
        typer.secho(f"stdout was: {ret.stdout!r}", err=True)
        typer.secho(f"stderr was: {ret.stderr!r}", err=True)
        raise typer.Exit(1)


def _log_popen_stdout(p):
    while p.poll() is None:
        line = p.stdout.readline().rstrip()
        if not line:
            continue
        bg, fg = None, None
        if match := LOG_PATTERN.match(line):
            bg, fg = LOG_LEVEL_MAP.get(match.groups()[0], (bg, fg))
        typer.secho(line, err=True, bg=bg, fg=fg)


if __name__ == "__main__":
    app()



""" The POOL XML File module provides a means to extract the GUID of a file or list
    of files by searching for an appropriate POOL XML Catalog in the specified directory.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import glob
import tarfile

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog
from DIRAC.Core.Utilities.List import uniqueElements
from DIRAC.Core.Utilities.File import makeGuid

__RCSID__ = "$Id$"

#############################################################################


def getGUID(fileNames, directory=""):
    """This function searches the directory for POOL XML catalog files and extracts the GUID.

    fileNames can be a string or a list, directory defaults to PWD.
    """

    if not directory:
        directory = os.getcwd()

    if not os.path.isdir(directory):
        return S_ERROR("%s is not a directory" % directory)

    if not isinstance(fileNames, list):
        fileNames = [fileNames]

    gLogger.verbose("Will look for POOL XML Catalog GUIDs in %s for %s" % (directory, ", ".join(fileNames)))

    finalCatList = _getPoolCatalogs(directory)

    # Create POOL catalog with final list of catalog files and extract GUIDs
    generated = []
    pfnGUIDs = {}
    catalog = PoolXMLCatalog(finalCatList)
    for fname in fileNames:
        guid = str(catalog.getGuidByPfn(fname))
        if not guid:
            guid = makeGuid(fname)
            generated.append(fname)

        pfnGUIDs[fname] = guid

    if not generated:
        gLogger.info("Found GUIDs from POOL XML Catalogue for all files: %s" % ", ".join(fileNames))
    else:
        gLogger.info("GUIDs not found from POOL XML Catalogue (and were generated) for: %s" % ", ".join(generated))

    result = S_OK(pfnGUIDs)
    result["directory"] = directory
    result["generated"] = generated
    return result


#############################################################################


def getType(fileNames, directory=""):
    """This function searches the directory for POOL XML catalog files and extracts the type of the pfn.

    fileNames can be a string or a list, directory defaults to PWD.
    """

    if not directory:
        directory = os.getcwd()

    if not os.path.isdir(directory):
        return S_ERROR("%s is not a directory" % directory)

    if not isinstance(fileNames, list):
        fileNames = [fileNames]

    gLogger.verbose("Will look for POOL XML Catalog file types in %s for %s" % (directory, ", ".join(fileNames)))

    finalCatList = _getPoolCatalogs(directory)

    # Create POOL catalog with final list of catalog files and extract GUIDs
    generated = []
    pfnTypes = {}
    catalog = PoolXMLCatalog(finalCatList)
    for fname in fileNames:
        typeFile = str(catalog.getTypeByPfn(fname))
        if not typeFile:
            typeFile = "ROOT_All"
            generated.append(fname)

        pfnTypes[fname] = typeFile

    if not generated:
        gLogger.info("Found Types from POOL XML Catalogue for all files: %s" % ", ".join(fileNames))
    else:
        gLogger.info("GUIDs not found from POOL XML Catalogue (and were generated) for: %s" % ", ".join(generated))

    result = S_OK(pfnTypes)
    result["directory"] = directory
    result["generated"] = generated
    return result


#############################################################################


def _getPoolCatalogs(directory=""):

    patterns = ["*.xml", "*.xml*gz"]
    omissions = [r"\.bak$"]  # to be ignored for production files

    # First obtain valid list of unpacked catalog files in directory
    poolCatalogList = []

    for pattern in patterns:
        fileList = glob.glob(os.path.join(directory, pattern))
        for fname in fileList:
            if fname.endswith(".bak"):
                gLogger.verbose("Ignoring BAK file: %s" % fname)
            elif tarfile.is_tarfile(fname):
                gLogger.debug("Unpacking catalog XML file %s" % (os.path.join(directory, fname)))
                with tarfile.open(os.path.join(directory, fname), "r") as tf:
                    for member in tf.getmembers():
                        tf.extract(member, directory)
                        poolCatalogList.append(os.path.join(directory, member.name))
            else:
                poolCatalogList.append(fname)

    poolCatalogList = uniqueElements(poolCatalogList)

    # Now have list of all XML files but some may not be Pool XML catalogs...
    finalCatList = []
    for possibleCat in poolCatalogList:
        try:
            _cat = PoolXMLCatalog(possibleCat)
            finalCatList.append(possibleCat)
        except Exception as x:
            gLogger.debug("Ignoring non-POOL catalogue file %s" % possibleCat)

    gLogger.debug("Final list of catalog files are: %s" % ", ".join(finalCatList))

    return finalCatList


#############################################################################
